"use strict";
function _export(target, all) {
    for(var name in all)Object.defineProperty(target, name, {
        get: all[name],
        enumerable: true
    });
}
_export(exports, {
    compile: ()=>compile,
    makeAceHighlighter: ()=>makeAceHighlighter,
    makeCss: ()=>makeCss
});
const _scopeManager = require("./ScopeManager");
const _js2Plist = require("plist2/js2plist");
const validGlobalAttributes = [
    "background_color",
    "color",
    "description",
    "file_extensions",
    "first_line_match",
    "folding_end_marker",
    "folding_start_marker",
    "name",
    "secondary_theme_background_color",
    "secondary_theme_color",
    "selected_theme",
    "textmate_compatible",
    "textmate_scope_override",
    "textmate_uuid", 
];
const validRegexAttribute = /^(?:\^|\\b|\\B)?(\(.*?\))+(?:\$|\\b|\\B)?$/;
const EOL_REGEX = "(^(?=.{0,1})(?:|))";
const NO_LOC = {
    start: {
        line: 1,
        column: 0,
        offset: 0
    },
    end: {
        line: 1,
        column: 0,
        offset: 0
    }
};
function isAttribute(node) {
    return node.type == "BasicAttribute" || node.type == "ArrayAttribute" || node.type == "RegexAttribute";
}
function getNamedNode(name, type, array) {
    return array.find((n)=>n.type == type && n["name"].name == name);
}
/**
 * @returns the number of subgroups in each capture group
 */ function getRegexGroups(regex) {
    var list = [];
    var captures = 1;
    var depth = 0;
    for(let i = 0; i < regex.length; i++){
        let c = regex[i];
        if (c == "\\") {
            i++;
        } else if (c == "(") {
            if (depth > 0) captures++;
            depth++;
        } else if (c == ")" && --depth == 0) {
            list.push(captures);
            captures = 1;
        }
    }
    return list;
}
function compile(ast, _options) {
    var options = Object.assign({
        targets: [],
        aceDefaultToken: true
    }, _options);
    //TODO: selective compilation
    //  a possible but slow solution is to have a separate pass for each grammar target
    const TEXTMATE = options.targets.includes("textmate");
    const ACE = options.targets.includes("ace");
    var tmGrammar = {
        name: "",
        scopeName: "",
        patterns: [
            {
                include: "#main"
            }
        ],
        repository: {}
    };
    var aceGrammar = {
        name: "",
        rules: {}
    };
    var errors = [];
    var scopes = new _scopeManager.ScopeManager();
    var inclusions = new Map();
    function checkRecursion(stack) {
        if (stack.length > 10) return [
            ...stack.slice(0, 5),
            "[...]"
        ];
        if (stack[0] == stack[1]) return stack;
        var targetName = stack[stack.length - 1];
        var target = inclusions.get(targetName);
        if (!target) return null;
        for (var subTarget of target){
            var _stack = [
                ...stack,
                subTarget
            ];
            if (stack.includes(subTarget)) return _stack;
            var res = checkRecursion(_stack);
            if (res) return res;
        }
        return null;
    }
    /**
     * Pushes an error or warning to the error list
     */ function err(message, location, fatal) {
        errors.push({
            message,
            location,
            fatal
        });
    }
    /**
     * Checks if the attribute exists and pushes an error otherwise
     * @param name The name of the attribute
     * @param type The attribute's type
     * @param node The attribute's parent node
     * @returns returns the attribute if it exists
     */ function mandatory(name, type, node) {
        var attr = getNamedNode(name, type, node.body);
        if (attr) return attr;
        else return err(`Could not find mandatory attribute '${name}'`, node.loc, true);
    }
    var styles = new Map();
    function getStyle(name, node) {
        var s = styles.get(name);
        if (s) return s;
        else err(`Could not find style '${name}'`, node.loc, true);
    }
    var currentCollection;
    var currentTMContext;
    var currentAceRule;
    /**
     * Creates an ace subrule
     * @param rule The name of the rule
     * @returns The name of the subrule
     */ function aceSubRule(rule) {
        rule = rule.replace(/__\d+$/, "");
        var i = 1;
        var newRule;
        do {
            newRule = rule + "__" + i++;
        }while (aceGrammar.rules[newRule])
        aceGrammar.rules[newRule] = [];
        return newRule;
    }
    var expandStack = [];
    /**
     * @returns the attribute value with expanded constants
     */ function expand(attr) {
        if (expandStack.includes(attr)) {
            let trace = "";
            for (let s of expandStack)trace += scopes.get(s.name.name)?.fullName + " → ";
            trace += scopes.get(attr.name.name)?.fullName;
            err(`Cyclic reference detected (${trace})`, attr.loc, true);
            return "";
        }
        expandStack.push(attr);
        var out;
        if (attr.type == "RegexAttribute" || attr.type == "BasicAttribute") {
            out = "";
            for (let val of attr.value.value){
                if (val.type == "Reference") {
                    var ref = scopes.get(val.name.name);
                    if (!ref) {
                        err(`Attribute '${val.name.name}' does not exist`, val.name.loc, true);
                        expandStack.pop();
                        return "";
                    }
                    if (ref.type == "ArrayAttribute") {
                        err(`Unexpected reference to array attribute ${ref.fullName}`, val.loc, true);
                        expandStack.pop();
                        return "";
                    }
                    // console.log("%s -> %s", scopes.getPath() + attr.name.name, ref.fullName || `'${ref}'`);
                    // console.log("->", ref.source);
                    if (ref.isExpanded) {
                        out += ref.expanded;
                    } else {
                        let exp = expand(ref.source);
                        scopes.get(ref.name).expanded = exp;
                        out += exp;
                    }
                } else {
                    out += val.value;
                }
            }
        } else {
            expandStack.pop();
            throw "Unexpected attribute type " + attr.type;
        }
        expandStack.pop();
        return out;
    }
    /**
     * Registers attributes into the current scope
     * @param body Node to find attributes in
     */ function registerAttributes(body) {
        // Attribute Declaration
        for (let node of body){
            if (isAttribute(node)) {
                let { name  } = node.name;
                if (scopes.current == scopes.global && !name.startsWith("__") && !validGlobalAttributes.includes(name)) {
                    throw new Error(`Unexpected attribute '${name}'`);
                }
                if (scopes.current.has(name)) err(`Duplicate attribute '${name}'`, node.loc, true);
                else scopes.set(name, node);
            }
        }
        // Attribute Expansion
        for (let node1 of body){
            if (isAttribute(node1)) {
                if (node1.type == "ArrayAttribute") {
                    let out = [];
                    for (let element of node1.value){
                        out.push(expand({
                            type: "BasicAttribute",
                            name: node1.name,
                            value: element,
                            loc: element.loc
                        }));
                    }
                    scopes.get(node1.name.name).expanded = out;
                } else scopes.get(node1.name.name).expanded = expand(node1);
            }
        }
    }
    //TODO: make optional in config
    function enforceDotPrefix(styleList) {
        for (let s of styleList){
            if (s[0][0] != ".") {
                err(`Type identifiers must start with a dot. Did you mean to write '.${s[0]}'?`, s[1].loc, false);
                s[0] = "." + s[0];
            }
        }
    }
    /**
     * Checks whether the value of a regex attribute is valid
     */ function validateRegex(regex, loc) {
        if (!validRegexAttribute.test(regex)) {
            err(`'${regex}' does not match the regular expression ${validRegexAttribute}`, loc, true);
            return false;
        } else {
            try {
                new RegExp(regex);
                return true;
            } catch (e) {
                err(e.message || "Invalid regular expression", loc, true);
                return false;
            }
        }
    }
    function validateRegexAndStyleList(regex, regexLoc, styleList, stylesLoc) {
        if (validateRegex(regex, regexLoc)) {
            // The number of styles must either
            // match the total number of regex capture groups eg. (a(b))(c(d(e))) -> 5
            // or the number of top groups eg. (a(b))(c(d(e))) -> 2
            let groups = getRegexGroups(regex);
            let totalGroups = groups.reduce((a, b)=>a + b);
            // If there are only styles for the top groups, populate the style list
            if (styleList.length == groups.length && groups.length != totalGroups) {
                let newList = [];
                for(let i = 0; i < styleList.length; i++){
                    for(let j = 0; j < groups[i]; j++)newList.push(styleList[i]);
                }
                // Replace the styles array but keeping reference
                styleList.splice(0, styleList.length, ...newList);
            } else if (styleList.length != totalGroups) {
                err("Number of styles and regexp groups doesn't match", stylesLoc, true);
            }
        //TODO: add this error check
        //All captures must be surrounded by brackets (for compatibility with Ace editor).
        //"\w([a-z][a-z0-9]*)" did not match regular expression  "^(?:\^|\\b|\\B)?(\(.*?\))+(?:\$|\\b|\\B)?$"
        }
        enforceDotPrefix(styleList);
    }
    /**
     * Traverses through an ast to generate grammars
     */ function traverse(node) {
        if (node.type == "Grammar") {
            registerAttributes(node.body);
            if (scopes.global.has("name")) {
                let name = scopes.global.get("name");
                tmGrammar.name = name.expanded;
                tmGrammar.scopeName = "source." + tmGrammar.name;
                aceGrammar.name = name.expanded;
                aceGrammar.name = aceGrammar.name[0].toUpperCase() + aceGrammar.name.toLowerCase().slice(1);
            } else err("Could not find mandatory attribute 'name'", NO_LOC, true);
            if (scopes.global.has("file_extensions")) {
                tmGrammar.fileTypes = scopes.global.get("file_extensions").expanded;
            }
            // Styles are registered first
            let stylesCollection = node.body.find((n)=>n.type == "Collection" && n.name.name == "styles");
            if (stylesCollection) traverse(stylesCollection);
            else err("Could not find mandatory collection 'styles'", NO_LOC, true);
            if (!node.body.find((n)=>n.type == "Collection" && n.name.name == "contexts")) err("Could not find mandatory collection 'contexts'", NO_LOC, true);
            for (let n of node.body){
                if (n.type == "Collection" && n.name.name == "contexts") {
                    // Register includes to detect cyclic references
                    for (let c of n.body){
                        if (c.type == "Object" && c.kind.name == "context" && c.name) {
                            let includeList = c.body.filter((n)=>n.type == "Object" && n.kind.name == "include");
                            let targets = includeList.map((n)=>n.body[0].value);
                            inclusions.set(c.name.name, targets.filter((t, i)=>targets.indexOf(t) == i));
                        }
                    }
                }
                if (!isAttribute(n) && n != stylesCollection) traverse(n);
            }
        } else if (node.type == "Collection") {
            if (scopes.current != scopes.global) return err(`Nested collections are not allowed`, node.loc, true);
            let { name: name1  } = node.name;
            if (name1 == "styles" || name1 == "contexts") {
                if (scopes.push(name1) === false) {
                    return err(`Duplicate collection '${name1}'`, node.name.loc, true);
                }
                for (let n1 of node.body){
                    if (name1 == "styles" && n1.type == "Object" && n1.name && n1.name.name[0] != ".") {
                        err(`Type identifiers must start with a dot. Did you mean to write '.${n1.name.name}'?`, n1.name.loc, false);
                        n1.name.name = "." + n1.name.name;
                    }
                    currentCollection = name1;
                    traverse(n1);
                }
                scopes.pop();
            } else err(`Unexpected collection '${name1}'`, node.name.loc, false);
        } else if (node.type == "Object") {
            let name2 = node.name?.name;
            let kind = node.kind.name;
            // If we are on the top/global scope
            if (!scopes.current.parent) {
                let errorMessage = `Unexpected global object ${name2 ? `'${name2}'` : ""}`;
                if (kind == "context" || kind == "pattern") errorMessage += ". Did you mean to place it inside collection 'contexts'?";
                if (kind == "pattern") errorMessage += ". Did you mean to place it inside object 'main'?";
                if (kind == "style") errorMessage += ". Did you mean to place it inside collection 'styles'?";
                err(errorMessage, node.loc, true);
                return;
            }
            if (node.name?.name == "__proto__") {
                err("Prototype pollution", node.name.loc, true);
                return;
            }
            let s = scopes.push(name2 || "?" + kind);
            if (s === false) {
                err(`Duplicate object '${name2}'`, node.loc, true);
                return;
            }
            registerAttributes(node.body);
            try {
                if (kind == "pattern") {
                    let regexNode = mandatory("regex", "RegexAttribute", node);
                    let stylesNode = mandatory("styles", "ArrayAttribute", node);
                    if (!regexNode || !stylesNode) return;
                    let regex = scopes.get(regexNode.name.name).expanded;
                    let styleList = scopes.get(stylesNode.name.name).expanded.map((v, i)=>[
                            v,
                            stylesNode.value[i]
                        ]);
                    validateRegexAndStyleList(regex, regexNode.loc, styleList, stylesNode.loc);
                    if (styleList.length == 1) {
                        let style = getStyle(styleList[0][0], styleList[0][1]);
                        if (style) {
                            currentTMContext.patterns.push({
                                name: style.textmate,
                                match: regex
                            });
                            aceGrammar.rules[currentAceRule].push({
                                token: style.ace,
                                regex: regex
                            });
                        }
                    } else {
                        let tmCaptures = {};
                        let aceTokens = [];
                        let i = 1;
                        for (let styleInfo of styleList){
                            let style1 = getStyle(styleInfo[0], styleInfo[1]);
                            if (!style1) continue;
                            tmCaptures[i++] = {
                                name: style1.textmate
                            };
                            aceTokens.push(style1.ace);
                        }
                        currentTMContext.patterns.push({
                            match: regex,
                            captures: tmCaptures
                        });
                        aceGrammar.rules[currentAceRule].push({
                            token: aceTokens,
                            regex: regex
                        });
                    }
                } else if (kind == "push") {
                    //TODO: implement push
                    err(`Missing implementation for 'push', try using 'inline_push' instead`, node.kind.loc, true);
                } else if (kind == "inline_push") {
                    let invalid = false;
                    let regexNode1 = mandatory("regex", "RegexAttribute", node);
                    let stylesNode1 = mandatory("styles", "ArrayAttribute", node);
                    let defaultStyleNode = getNamedNode("default_style", "BasicAttribute", node.body);
                    if (!node.body.find((n)=>n.type == "Object")) {
                        err("Please supply one or more inline list items", node.loc, true);
                        invalid = true;
                    }
                    let pop = node.body.filter((n)=>n.type == "Object" && (n.kind.name == "pop" || n.kind.name == "eol_pop"));
                    if (pop.length == 0) {
                        err("missing pop object", node.loc, true);
                        invalid = true;
                    }
                    if (pop.length > 1) {
                        err("too many pop objects", {
                            start: pop[1].loc.start,
                            end: pop[pop.length - 1].loc.end
                        }, true);
                        invalid = true;
                    }
                    let hasNonPop = node.body.find((n)=>n.type == "Object" && n.kind.name != "pop" && n.kind.name != "eol_pop");
                    if (hasNonPop && defaultStyleNode) {
                        err("Cannot set a default Style if a non-pop rule is provided or included (textmate mode)", defaultStyleNode.loc, true);
                        invalid = true;
                    }
                    if (!hasNonPop && !defaultStyleNode) {
                        err("Please supply a default Style for this rule (textmate mode)", node.loc, true);
                        invalid = true;
                    }
                    if (regexNode1 && stylesNode1 && !invalid) {
                        let regex1 = scopes.get(regexNode1.name.name).expanded;
                        let styleList1 = scopes.get(stylesNode1.name.name).expanded.map((v, i)=>[
                                v,
                                stylesNode1.value[i]
                            ]);
                        validateRegexAndStyleList(regex1, regexNode1.loc, styleList1, stylesNode1.loc);
                        let defaultStyle = defaultStyleNode && scopes.get(defaultStyleNode.name.name).expanded;
                        defaultStyleNode && enforceDotPrefix([
                            [
                                defaultStyle,
                                defaultStyleNode.value
                            ]
                        ]);
                        let tmPattern = {
                            begin: regex1,
                            beginCaptures: {},
                            end: undefined,
                            endCaptures: {},
                            patterns: []
                        };
                        let acePattern = {
                            token: [],
                            regex: regex1,
                            push: aceSubRule(currentAceRule)
                        };
                        currentTMContext.patterns.push(tmPattern);
                        aceGrammar.rules[currentAceRule].push(acePattern);
                        let i1 = 1;
                        for (let styleInfo1 of styleList1){
                            let style2 = getStyle(styleInfo1[0], styleInfo1[1]);
                            if (!style2) continue;
                            tmPattern.beginCaptures[i1++] = {
                                name: style2.textmate
                            };
                            acePattern.token.push(style2.ace);
                        }
                        if (acePattern.token.length == 1) acePattern.token = acePattern.token[0];
                        let _currentTMContext = currentTMContext, _currentAceContext = currentAceRule;
                        currentTMContext = tmPattern;
                        currentAceRule = acePattern.push;
                        for (let n2 of node.body){
                            if (!isAttribute(n2)) traverse(n2);
                        }
                        currentAceRule = _currentAceContext;
                        currentTMContext = _currentTMContext;
                        if (tmPattern.patterns.length == 0) delete tmPattern.patterns;
                        if (defaultStyle) {
                            let style3 = getStyle(defaultStyle, defaultStyleNode.value);
                            if (!style3) return; //TODO: error: no styles defined
                            tmPattern.contentName = style3.textmate;
                            aceGrammar.rules[acePattern.push].push({
                                defaultToken: style3.ace
                            });
                        }
                    }
                } else if (kind == "pop" && currentTMContext["beginCaptures"]) {
                    let regexNode2 = mandatory("regex", "RegexAttribute", node);
                    let stylesNode2 = mandatory("styles", "ArrayAttribute", node);
                    if (!regexNode2 || !stylesNode2) return;
                    var regex2 = scopes.get(regexNode2.name.name).expanded;
                    let styleList2 = scopes.get(stylesNode2.name.name).expanded.map((v, i)=>[
                            v,
                            stylesNode2.value[i]
                        ]);
                    validateRegexAndStyleList(regex2, regexNode2.loc, styleList2, stylesNode2.loc);
                    let tmPattern1 = currentTMContext;
                    let acePattern1 = {
                        token: [],
                        regex: regex2,
                        next: "pop"
                    };
                    aceGrammar.rules[currentAceRule].push(acePattern1);
                    tmPattern1.end = regex2;
                    let i2 = 1;
                    for (let styleInfo2 of styleList2){
                        let style4 = getStyle(styleInfo2[0], styleInfo2[1]);
                        if (!style4) continue;
                        tmPattern1.endCaptures[i2++] = {
                            name: style4.textmate
                        };
                        acePattern1.token.push(style4.ace);
                    }
                    if (acePattern1.token.length == 1) acePattern1.token = acePattern1.token[0];
                } else if (kind == "eol_pop") {
                    let anyStyle = styles.entries().next().value;
                    currentTMContext["end"] = EOL_REGEX;
                    currentTMContext["endCaptures"]["1"] = {
                        name: anyStyle.textmate
                    };
                    aceGrammar.rules[currentAceRule].push({
                        token: anyStyle.ace,
                        regex: EOL_REGEX,
                        next: "pop"
                    });
                } else if (kind == "include") {
                    let current = scopes.current.parent.name;
                    let target = node.body[0].value;
                    if (!inclusions.has(target)) {
                        err(`Unknown context '${target}'`, node.body[0].loc, true);
                        return;
                    }
                    // console.log(scopes.current);
                    // if (targets.indexOf(n) != i) {
                    //     err("Redundant include", includeList[i].loc, true); //TODO: make this a warning instead
                    // }
                    // Test for circular inclusions
                    let result = checkRecursion([
                        current,
                        target
                    ]);
                    if (result && current == result[result.length - 1]) {
                        var trace = result.join(" → ");
                        err(`Cyclic inclusion detected (${trace})`, node.loc, true);
                    } else {
                        currentTMContext.patterns.push({
                            include: "#" + target
                        });
                        //NOTE: Include doesn't exist in ace, we're just adding this to be replaced later
                        aceGrammar.rules[currentAceRule].push({
                            include: target
                        });
                    }
                } else if (kind == "embed") {
                    let textmateTarget = mandatory("textmate", "BasicAttribute", node);
                    let aceTarget = mandatory("ace", "BasicAttribute", node);
                    console.log({
                        textmateTarget,
                        aceTarget
                    });
                } else if (kind == "context") {
                    if (currentCollection != "contexts") {
                        err(`Unexpected object type ${kind}. Did you mean to place it in collection 'contexts'?`, node.kind.loc, true);
                        return;
                    }
                    if (!name2) return err(`Unnamed context`, node.loc, false);
                    //@ts-expect-error
                    tmGrammar.repository[name2] = currentTMContext = {
                        patterns: []
                    };
                    aceGrammar.rules[currentAceRule = name2] = [];
                    for (let n3 of node.body){
                        if (!isAttribute(n3)) traverse(n3);
                    }
                    currentTMContext = null;
                } else if (kind == "style") {
                    if (currentCollection == "styles") {
                        if (!name2) return err(`Unnamed style`, node.loc, false);
                        let style5 = {
                            name: name2,
                            textmate: "text." + tmGrammar.name,
                            ace: "text." + tmGrammar.name
                        };
                        let color = s.get("color")?.expanded;
                        if (color) style5.color = color;
                        let textmate = s.get("textmate_scope")?.expanded;
                        let ace = s.get("ace_scope")?.expanded;
                        if (textmate) style5.textmate = textmate + "." + tmGrammar.name;
                        if (ace) style5.ace = ace;
                        if (textmate && !ace) style5.ace = textmate;
                        if (!textmate && ace) style5.textmate = style5.ace;
                        styles.set(name2, style5);
                    } else err(`Unexpected object type ${kind}. Did you mean to place it in collection 'styles'?`, node.kind.loc, true);
                } else {
                    err(`Unexpected object type ${kind}`, node.kind.loc, true);
                }
            } finally{
                // Ensures we pop the scope even if we return after an erorr
                scopes.pop();
            }
        } else throw new Error("Unexpected node type " + node["type"]);
    }
    traverse(ast);
    if (!errors.find((e)=>e.fatal)) {
        // Expand includes
        for(let r in aceGrammar.rules){
            //TODO: prevent circular includes
            let rule = aceGrammar.rules[r];
            let pos;
            //TODO: improve performance by finding from last index + added items
            while((pos = rule.findIndex((p)=>"include" in p)) > -1){
                let target = rule[pos]["include"];
                // console.log(r, target, aceGrammar.rules[target]);
                rule.splice(pos, 1, ...aceGrammar.rules[target]);
            }
        }
        // Rename main to start
        aceGrammar.rules.start = aceGrammar.rules.main || [];
        delete aceGrammar.rules.main;
        for(let r1 in aceGrammar.rules){
            // Remove unused rules
            if (r1 != "start" && !r1.match(/__\d+$/)) delete aceGrammar.rules[r1];
            else if (options.aceDefaultToken) {
                let rr = aceGrammar.rules[r1];
                if (rr.length > 0 && "defaultToken" in rr[rr.length - 1]) rr.push({
                    defaultToken: "text"
                });
            }
        }
    }
    return {
        errors,
        textmate: tmGrammar,
        ace: aceGrammar,
        makeTextmateXML: ()=>(0, _js2Plist.default)(tmGrammar),
        makeAceHighlighter: (options)=>makeAceHighlighter(aceGrammar, options),
        makeCss: ()=>makeCss(styles),
        styles,
        global: scopes.global,
        ast
    };
}
function makeAceHighlighter(aceGrammar, options) {
    var { name , rules  } = aceGrammar;
    return `define(function(require, exports, module) {
"use strict";
var oop = require("../lib/oop");
var TextHighlightRules = require("./text_highlight_rules").TextHighlightRules;
/* --------------------- START ----------------------------- */
var ${name}HighlightRules = function() {
this.$rules = ${JSON.stringify(rules, null, options?.spacing)};
this.normalizeRules();
};
/* ------------------------ END ------------------------------ */
oop.inherits(${name}HighlightRules, TextHighlightRules);
exports.${name}HighlightRules = ${name}HighlightRules;
});`;
}
function makeCss(styles) {
    //TODO: make css
    return ``;
}

//# sourceMappingURL=compiler.js.map